@@ -68,7 +68,7 @@ namespace simgear
         }
     protected:
         HTTPDirectory* _directory;
-        size_t _contentSize;
+        size_t _contentSize = 0;
     };
 
     typedef SGSharedPtr<HTTPRepoGetRequest> RepoRequestPtr;
@@ -756,7 +756,12 @@ size_t HTTPRepository::bytesToDownload() const
     }
 
     for (r = _d->activeRequests.begin(); r != _d->activeRequests.end(); ++r) {
-        result += (*r)->contentSize() - (*r)->responseBytesReceived();
+        if ((*r)->contentSize() > 0) {
+            // Content size for root dirindex of a repository is zero,
+            // and returing a negative value breaks everyting, so just ignore
+            // it
+            result += (*r)->contentSize() - (*r)->responseBytesReceived();
+        }
     }
 
     return result;
@@ -1239,7 +1244,7 @@ HTTPRepository::failure() const
 
     void HTTPRepoPrivate::failedToGetRootIndex(HTTPRepository::ResultCode st)
     {
-        SG_LOG(SG_TERRASYNC, SG_WARN, "Failed to get root of repo:" << baseUrl);
+        SG_LOG(SG_TERRASYNC, SG_WARN, "Failed to get root of repo:" << baseUrl << " " << st);
         status = st;
     }
 
@@ -1288,7 +1293,6 @@ HTTPRepository::failure() const
 
         // maybe there was nothing to do
         if (activeRequests.empty()) {
-            status = HTTPRepository::REPO_NO_ERROR;
             isUpdating = false;
         }
     }
